/*
 * @Author       : kaikai.hou
 * @Email        : kaikai.hou@downtown8.com
 * @Description  : Balabala
 * @Date         : 2020-02-10 16:11:59
 * @LastEditors  : kaikai.hou
 * @LastEditTime : 2020-02-11 19:33:16
 */

interface GroupByParam {
  list?: Array<any>; //数组
  array?: Array<any>;
  key: any; //分组的key
  fun: Function; // 处理函数
}

interface AbsJoinParam {
  list: Array<any>;
  list2: Array<any>;
  fun?: Function;
  keys?: Array<string>;
}

interface JoinOpt {
  list: Array<any>;
  list2: Array<any>;
  /**
   * 两边都有的死后处理的函数
   * function(array1,array2,e)
   */
  fun: Function;
  /**
   * 对应的key
   */
  key: Function | string | Array<string>;
  /**
   * 默认和key相同
   */
  key2?: Function | string | Array<string>;
  /**
   * 只有数组1有的时候处理函数
   * function(array,e)
   */
  onlyOne?: Function;
  /**
   * 只有数组2有的时候处理函数
   * function(array,e)
   */
  onlyTwo?: Function;
}

/**
从一个元素中取值
也支持有function
*/
function get(obj: object, key: any) {
  if (key == null) return obj;
  if (key instanceof Function) {
    return key(obj);
  }
  const ret = [];
  if (key instanceof Array) {
    for (const k of key) {
      ret.push(obj[k]);
    }
    return ret.join('___');
  } else {
    return obj[key];
  }
}

export class ArrayUtil {
  static keys(array: any[]) {
    return array.join('___');
  }

  /**
   * 求各个key不重复的数量
   * @param array
   * @param keys
   */
  static countObj(array: any[], keys: string[]) {
    const keyMap: any = {};
    const ret: any = {};
    for (const key of keys) {
      ret[key] = 0;
      keyMap[key] = {};
    }
    for (const row of array) {
      for (const key of keys) {
        const value = row[key];
        if (value != null) {
          if (keyMap[key][value] == null) {
            keyMap[key][value] = true;
            ret[key]++;
          }
        }
      }
    }
    return ret;
  }

  /**
   * 得到一个object，这个object 含有所有的key值
   * @param array
   * @param keys
   */
  static sumObj(array: any[], keys: string[]) {
    const ret: any = {};
    for (const row of array) {
      for (const key of keys) {
        if (ret[key] == null) {
          ret[key] = row[key];
        } else {
          if (row[key] != null) ret[key] += row[key];
        }
      }
    }
    return ret;
  }

  static sum(
    array: Array<any>,
    key?: Array<string> | Function | string,
  ): number {
    let ret = 0;
    if (!array) return 0;
    const len = array.length;
    for (let i = 0; i < len; i++) {
      const obj = array[i];
      if (key) {
        const val = get(obj, key);
        if (val != null) {
          ret += val;
        }
      } else {
        ret += obj;
      }
    }
    return ret;
  }
  static distinctByKey(array, keys: string | Function | Array<string>) {
    const map = ArrayUtil.toMapByKey(array, keys);
    const list = [];
    for (const e in map) {
      list.push(map[e]);
    }
    return list;
  }
  /**
   * 根据这个产生一个key
   * @param obj
   * @param keys
   */
  static get(obj, keys: string | Function | Array<string>) {
    return get(obj, keys);
  }
  /**
   * 排序
   * @param array 排序数组
   * @param opts string|{order:'name',desc:'desc'}
   *
   *
   */
  static order(array, opts: any): Array<any> {
    if (!(opts instanceof Array)) {
      if (opts.order == null) {
        opts = {
          order: opts,
        };
      }
      opts = [opts];
    }

    function sort(obj1, obj2) {
      function createFun(opt) {
        if (opt.fun) {
          return opt.fun;
        }
        return function (obj1, obj2) {
          const order = opt.order;
          const desc = opt.desc;
          let ret = 0;
          let d1 = get(obj1, order);
          let d2 = get(obj2, order);
          if (d1 == null && d2 == null) {
            ret = 0;
          } else if (d1 == null && d2 != null) {
            ret = -1;
          } else if (d1 != null && d2 == null) {
            ret = 1;
          } else {
            if (d1 instanceof Date) {
              d1 = d1.getTime();
            }
            if (d2 instanceof Date) {
              d2 = d2.getTime();
            }
            if (d1 < d2) {
              ret = -1;
            } else if (d1 > d2) {
              ret = 1;
            }
          }

          if (desc == 'desc') {
            ret = ret * -1;
          }

          return ret;
        };
      }
      const funs = ArrayUtil.parse(opts, createFun);
      for (let i = 0; i < funs.length; i++) {
        const ret = funs[i](obj1, obj2);
        if (ret != 0 || i == funs.length - 1) {
          return ret;
        }
      }
    }
    array.sort(sort);
    return array;
  }
  /**
   *
   * @param opt
   * @returns
   */
  static async groupBySync(opt: GroupByParam): Promise<any[]> {
    if (opt == null) {
      return null;
    }
    if (opt.key == null) {
      throw new Error('opt不能为null');
    }
    const list = opt.list || opt.array;
    if (list == null) {
      return null;
    }
    let fun = opt.fun;
    if (fun == null) {
      fun = (array) => array;
    }
    let key = opt.key;
    if (key == null) {
      key = 'id';
    }
    const map = ArrayUtil.toMapArray(list, key);
    const ret = [];
    for (const e in map) {
      const row = await fun(map[e], e);
      if (row != null) {
        if (row instanceof Array) {
          ret.push(...row);
        } else {
          ret.push(row);
        }
      }
    }
    return ret;
  }
  /**
     对数组做group by操作
	opt:{
		list:list, //数组
		key:key, //分组的key
		fun:fun(list,e) // 处理函数
	}
	*/
  static groupBy(opt: GroupByParam): Array<any> {
    if (opt == null) {
      return null;
    }
    if (opt.key == null) {
      throw new Error('opt不能为null');
    }
    const list = opt.list || opt.array;
    if (list == null) {
      return null;
    }
    let fun = opt.fun;
    if (fun == null) {
      fun = (array) => array;
    }
    let key = opt.key;
    if (key == null) {
      key = 'id';
    }
    const map = ArrayUtil.toMapArray(list, key);
    const ret = [];
    for (const e in map) {
      const row = fun(map[e], e);
      if (row != null) {
        if (row instanceof Array) {
          ret.push(...row);
        } else {
          ret.push(row);
        }
      }
    }
    return ret;
  }
  /**
	 * 将 通过比较key将两个数组按and的方式and
	   
        array1:数组1
        array2:数组2
        key1 
        key2
	 */
  static andByKey(array1: Array<any>, array2: Array<any>, key1?, key2?) {
    if (key1 == null) {
      key1 = (row) => row;
    }

    if (key2 == null) {
      key2 = key1;
    }

    const map = ArrayUtil.toMapByKey(array1, key1);
    const retMap = {};
    const ret = [];
    for (let i = 0; i < array2.length; i++) {
      const item = array2[i];
      const keyValue = get(item, key2);
      if (map[keyValue]) {
        retMap[keyValue] = map[keyValue];
      }
    }
    for (const e in retMap) {
      ret.push(retMap[e]);
    }
    return ret;
  }
  static inArray(array: Array<any>, obj: any): boolean {
    if (obj == null) return false;
    if (!Array.isArray(obj)) {
      const len = array.length;
      for (let i = 0; i < len; i++) {
        if (obj == array[i]) {
          return true;
        }
      }
      return false;
    } else {
      obj = ArrayUtil.distinct(obj);
      const list = ArrayUtil.and(array, obj);
      return list.length == obj.length;
    }
  }
  static and(array1, array2): Array<any> {
    const map = ArrayUtil.toMap(array1);
    const retMap = {};
    const ret = [];
    for (let i = 0; i < array2.length; i++) {
      const val = array2[i];
      if (map[val]) {
        retMap[val] = val;
      }
    }
    for (const e in retMap) {
      ret.push(retMap[e]);
    }
    return ret;
  }
  static distinct(array: Array<any>): Array<any> {
    if (array == null || array.length == 0) return array;
    const map = {};
    for (const row of array) {
      map[row] = row;
    }
    const ret = [];
    for (const e in map) {
      ret.push(map[e]);
    }
    return ret;
  }

  /**
	建议使用
	 * 根据key 转化为map
	 * @param array
	 * @param key
	 */
  static toMapByKey(array: any[], key: any, fun?: Function | string) {
    if (fun == null) {
      fun = (data) => {
        return data;
      };
    }
    const map = {};
    if (!key) key = 'id';
    if (array) {
      const len = array.length;
      for (let i = 0; i < len; i++) {
        const data = array[i];
        const mapKey = get(data, key);
        map[mapKey] = get(data, fun);
      }
    }
    return map;
  }

  /**
   * @description 将一个list按key分组，放在map中
   */
  static toMapArray(list: any[], key: any, fun?: Function) {
    const ret = {};
    if (list != null && key != null) {
      for (let i = 0; i < list.length; i++) {
        const data = list[i];
        const val = get(data, key);
        const mapData = get(data, fun);
        if (val != null && mapData != null) {
          let array = ret[val];
          if (array == null) {
            array = [];
            ret[val] = array;
          }
          array.push(mapData);
        }
      }
    }
    return ret;
  }

  static copy(destArray, srcArray, maxLen) {
    if (destArray == null) return [];
    if (srcArray == null) return destArray;
    let len = srcArray.length;
    if (maxLen != null && maxLen >= 0 && maxLen < len) {
      len = maxLen;
    }
    for (let i = 0; i < len; i++) {
      destArray.push(srcArray[i]);
    }
    return destArray;
  }

  static toMap(array) {
    const obj = {};
    if (array) {
      if (!(array instanceof Array)) {
        array = [array];
      }
      for (let i = 0; i < array.length; i++) {
        obj[array[i]] = true;
      }
    }
    return obj;
  }

  /**
   * 两个数组求not_in,根据key 来 进行
   * @param array1
   * @param array2
   * @param key
   * @param key2
   */
  static notInByKey(array1: any[], array2: any[], key?, key2?) {
    if (key == null) {
      return ArrayUtil.notIn(array1, array2);
    }
    if (key2 == null) {
      key2 = key;
    }
    const map = ArrayUtil.toMapByKey(array2, key2);
    const retMap = {};
    for (let i = 0; i < array1.length; i++) {
      const val = array1[i];
      const valKey = get(val, key);
      if (map[valKey] == null) {
        retMap[valKey] = val;
      }
    }
    const array = [];
    for (const e in retMap) {
      array.push(retMap[e]);
    }
    return array;
  }
  /**
   * 查找最大的
   * @param array
   * @param fun
   * @returns
   */
  static findMax(array: any[], fun: string | ((row: any) => any)) {
    let ret = null;
    let val = null;
    for (const row of array) {
      if (ret == null) {
        ret = row;
        val = get(row, fun);
      } else {
        const rowVal = get(row, fun);
        if (rowVal > val) {
          val = rowVal;
          ret = row;
        }
      }
    }
    return ret;
  }

  /**
   * 查找最大的
   * @param array
   * @param fun
   * @returns
   */
  static findMin(array: any[], fun: string | ((row: any) => any)) {
    let ret = null;
    let val = null;
    for (const row of array) {
      if (ret == null) {
        ret = row;
        val = get(row, fun);
      } else {
        const rowVal = get(row, fun);
        if (rowVal < val) {
          val = rowVal;
          ret = row;
        }
      }
    }
    return ret;
  }

  /**
	第二个参数可以是function、map、array
	*/
  static filter(array, fun): Array<any> {
    const ret = [];
    if (!(array instanceof Array)) {
      array = [array];
    }
    for (let i = 0; i < array.length; i++) {
      if (fun(array[i], i)) {
        ret.push(array[i]);
      }
    }

    return ret;
  }
  /**
   * 过滤出一个
   * @param array
   * @param fun
   */
  static filterOne(array, fun): any {
    return ArrayUtil.filter(array, fun)[0];
  }
  /**
   * 找出在array1 不在array2的数据
   */
  static notIn(array1, array2): Array<any> {
    const map = ArrayUtil.toMap(array2);

    const retMap = {};
    for (let i = 0; i < array1.length; i++) {
      const val = array1[i];
      if (map[val] == null) {
        retMap[val] = val;
      }
    }
    const ret = [];
    for (const e in retMap) {
      ret.push(retMap[e]);
    }
    return ret;
  }

  /**
   * 将一个数组转换成另外一个数组
   * 和系统的map 类似
   * 之处传入一个asyn 函数，支持函数返回值为数组，将每次运行返回的数组组合成一个数组
   * @param list
   * @param fun
   */
  static async map(list: Array<any>, fun) {
    const ret = [];
    for (let i = 0; i < list.length; i++) {
      const funRet = await fun(list[i], i);
      if (funRet != null) {
        if (funRet instanceof Array) {
          ArrayUtil.addAll(ret, funRet);
        } else {
          ret.push(funRet);
        }
      }
    }
    return ret;
  }

  static parse(array, fun?: Function) {
    if (!array) return [];
    if (!(array instanceof Array)) {
      array = [array];
    }
    return ArrayUtil._parseArray(array, fun);
  }

  /**
   * 根据fun转化array后返回
   * @param array
   * @param fun
   * @returns
   */
  private static _parseArray(array: any[], fun: Function) {
    if (!array) return [];
    if (!fun) {
      return array;
    }
    const ret = [];
    for (let i = 0; i < array.length; i++) {
      const obj = fun(array[i], i);
      if (obj != null) {
        ret.push(obj);
      }
    }
    return ret;
  }

  static addAll(array1, array2) {
    if (array1 == null) array1 = [];
    if (array2 == null) return array1;
    if (!(array2 instanceof Array)) {
      array1.push(array2);
    } else {
      for (let i = 0; i < array2.length; i++) {
        array1.push(array2[i]);
      }
    }
    return array1;
  }

  static toArray(array, key: string | Function | Array<string>) {
    if (array == null) return [];
    if (!(array instanceof Array)) array = [array];
    if (!key) key = 'id';
    const ret = [];
    if (array) {
      for (let i = 0; i < array.length; i++) {
        const obj = get(array[i], key);
        if (obj != null) {
          ret.push(obj);
        }
      }
    }
    return ret;
  }

  static isSame(array1: Array<any>, array2: Array<any>): boolean {
    function _isSame(a: any, b: any) {
      const aProps = Object.getOwnPropertyNames(a);
      const bProps = Object.getOwnPropertyNames(b);

      if (aProps.length != bProps.length) {
        return false;
      }

      for (let i = 0; i < aProps.length; i++) {
        const propName = aProps[i];
        if (a[propName] !== b[propName]) {
          return false;
        }
      }

      return true;
    }
    if (array1 == null || array2 == null) {
      return false;
    }
    if (array1.length != array2.length) return false;
    for (let i = 0; i < array1.length; i++) {
      if (!_isSame(array1[i], array2[i])) {
        return false;
      }
    }

    return true;
  }

  static orByKey(
    array1,
    array2,
    key1?: string | Array<string> | Function,
    key2?: string | Array<string> | Function,
  ) {
    if (key1 == null) {
      key1 = (row) => row;
    }
    if (key2 == null) key2 = key1;
    const map = ArrayUtil.toMapByKey(array1, key1);

    for (let i = 0; i < array2.length; i++) {
      const data = array2[i];
      const key = get(data, key2);
      if (map[key] == null) {
        map[key] = data;
      }
    }
    const array = [];
    for (const e in map) {
      array.push(map[e]);
    }
    return array;
  }

  /**
	两个list进行join操作,
	类似数据库中inner join
	{
		list:[],
		list2:[],
		fun:function(data,data1){
			return data
		},
		key:function(){ //可以是function 也可以是key 在list中必须唯一

		},
		key2:function(){

		},
		onlyOne:function(data){ //只有第一个数组有的情况

		},
		onlyTwo:function(data){ //只存在2，不存在1

		}

	}
	*/
  static join(opt: JoinOpt): Array<any> {
    const list = opt.list;
    const list2 = opt.list2;
    if (list == null || list2 == null) {
      return [];
    }
    let key = opt.key;
    let key2 = opt.key2;
    let fun = opt.fun;
    if (key == null) {
      key = 'id';
    }
    if (key2 == null) {
      key2 = key;
    }
    if (fun == null) {
      fun = function (data) {
        return data;
      };
    }
    const onlyOne = opt.onlyOne;
    const onlyTwo = opt.onlyTwo;
    const map = ArrayUtil.toMapByKey(list, key);
    const ret = [];
    for (var i = 0; i < list2.length; i++) {
      var data2 = list2[i];
      var keyValue = get(data2, key2);
      var data1 = map[keyValue];
      if (data1 != null) {
        var row = fun(data1, data2);
        if (row != null) {
          ret.push(row);
        }
      } else {
        if (onlyTwo != null) {
          var row = onlyTwo(data2);
          if (row != null) {
            ret.push(row);
          }
        }
      }
    }
    if (onlyOne != null) {
      const map2 = ArrayUtil.toMapByKey(list2, key2);
      for (var i = 0; i < list.length; i++) {
        var data1 = list[i];
        var keyValue = get(data1, key);
        var data2 = map2[keyValue];
        if (data2 == null) {
          var row = onlyOne(data1);
          if (row != null) {
            ret.push(row);
          }
        }
      }
    }
    return ret;
  }

  /**
	两个list进行map操作,
	joinArray 方法和join类似，
	join 是一对一的关系，
	joinArray 是一对多的关系
	list2 是多条数据
	类似数据库中inner join
	{
		list:[],
		list2:[],
		fun:function(data,array){
			return data
		},
		key:function(){ //可以是function 也可以是key 在list中必须唯一

		},
		key2:function(){

		},
		onlyOne:function(data){ //只有第一个数组有的情况

		},
		onlyTwo:funtion(data){ //只存在2，不存在1

		}

	}
	*/
  static joinArray(opt: JoinOpt) {
    let list = opt.list;
    let list2 = opt.list2;
    if (list == null && list2 == null) return [];
    if (list == null) {
      list = [];
    }
    if (list2 == null) list2 = [];
    let key = opt.key;
    let key2 = opt.key2;
    let fun = opt.fun;
    if (key == null) {
      key = 'id';
    }
    if (key2 == null) {
      key2 = key;
    }
    if (fun == null) {
      fun = function (data) {
        return data;
      };
    }
    const map = ArrayUtil.toMapByKey(list, key);
    const map2 = ArrayUtil.toMapArray(list2, key2);
    const ret = [];
    for (var e in map) {
      const data = map[e];
      const array = map2[e];
      if (array != null) {
        var row = fun(data, array);
        ArrayUtil.addAll(ret, row);
      } else {
        if (opt.onlyOne) {
          var row = opt.onlyOne(data);
          ArrayUtil.addAll(ret, row);
        }
      }
    }
    if (opt.onlyTwo) {
      for (var e in map2) {
        if (map[e] == null) {
          var row = opt.onlyTwo(map2[e]);
          ArrayUtil.addAll(ret, row);
        }
      }
    }
    return ret;
  }

  /**
	两个list进行join操作,
	类似数据库中inner join
	和join不同是 两个数组都存在多个
	{
		list:[],
		list2:[],
		fun:function(array1,array2,key){
			return data
		},
		key:function(){ //可以是function 也可以是key 在list中必须唯一

		},
		key2:function(){

		},
		onlyOne:function(array,key){ //只有第一个数组有的情况

		},
		onlyTwo:funtion(data){ //只存在2，不存在1

		}

	}
	*/
  static joinMany(opt: JoinOpt) {
    const list = opt.list;
    const list2 = opt.list2;
    const onlyOne = opt.onlyOne;
    const onlyTwo = opt.onlyTwo;
    if (list == null || list2 == null) {
      return [];
    }
    let key = opt.key;
    let key2 = opt.key2;
    let fun = opt.fun;
    if (key == null) {
      key = 'id';
    }
    if (key2 == null) {
      key2 = key;
    }
    if (fun == null) {
      fun = function (data) {
        return data;
      };
    }

    const map = ArrayUtil.toMapArray(list, key);
    const map2 = ArrayUtil.toMapArray(list2, key2);
    const ret = [];
    for (var e in map2) {
      var array2 = map2[e];
      var array = map[e];
      if (array != null) {
        ArrayUtil.addAll(ret, fun(array, array2, e));
      } else {
        if (onlyTwo) ArrayUtil.addAll(ret, onlyTwo(array2, e));
      }
    }
    if (onlyOne != null) {
      for (var e in map) {
        var array = map[e];
        var array2 = map2[e];
        if (array2 == null) {
          ArrayUtil.addAll(ret, onlyOne(array, e));
        }
      }
    }
    return ret;
  }
  /**
   * 根据字段取得某个值的数组 并去重
   * @param array
   * @param key
   */
  static toArrayDis(
    array: Array<any>,
    key: string | Function | Array<string>,
  ): Array<any> {
    if (array == null) return null;
    const set = new Set();
    for (const row of array) {
      const val = get(row, key);
      if (val != null) set.add(val);
    }
    return [...set];
  }
  /**
   * 将一个mapArray 转成map
   * @param map
   */
  static mapArray2Array(map) {
    const array = [];
    for (const e in map) {
      ArrayUtil.addAll(array, map[e]);
    }
    return array;
  }

  /**
   * 判断两个数组是否相等
   * @param array1
   * @param array2
   * @param key
   */
  static isSameByKey(
    array1: Array<any>,
    array2: Array<any>,
    key: string | Array<string> | Function,
  ) {
    if (array1 == null && array2 == null) return true;
    if (array1 == null || array2 == null) return false;
    if (array1.length != array2.length) return false;
    for (let i = 0; i < array1.length; i++) {
      if (get(array1[i], key) != get(array2[i], key)) return false;
    }
    return true;
  }

  /**
	将两个数组笛卡尔相乘，返回一个新数组
	新数组的长度 是两个数组的长度乘积
	新数组的内容是fun函数的执行结果，null不会保存到结果里面
	opt{
		list:[],
		list2:[],
		fun:function(data,data2 ,i,t) //i和t 是数组的索引
	}
	*/
  static absJoin(opt: AbsJoinParam) {
    const array = [];
    const list = opt.list;

    const list2 = opt.list2;
    if (opt.fun == null) {
      const keys = opt.keys;
      opt.fun = function (data1, data2) {
        const obj = {};
        obj[keys[0]] = data1[keys[0]];
        obj[keys[1]] = data2[keys[1]];
        return obj;
      };
    }
    for (let i = 0; i < list.length; i++) {
      for (let t = 0; t < list2.length; t++) {
        const row = opt.fun(list[i], list2[t], i, t);
        if (row != null) {
          array.push(row);
        }
      }
    }
    return array;
  }

  /**
   * 只取key指定的值
   * @param list
   * @param key
   */
  static onlyKeys(list: any[], key) {
    if (key == null) return null;
    if (!(key instanceof Array)) {
      key = [key];
    }
    const array = [];
    for (const data of list) {
      const obj = {};
      for (const dataKey of key) {
        obj[dataKey] = data[dataKey];
      }
      array.push(obj);
    }
    return array;
  }
}
