import {
  CallHandler,
  ExecutionContext,
  Injectable,
  NestInterceptor,
} from '@nestjs/common';

import { Observable } from 'rxjs';
import { tap } from 'rxjs/operators';
import { LoggerService } from './logger.service';
import { IGNORE_URL } from '@/common/lib/constant';
@Injectable()
export class LoggingInterceptor implements NestInterceptor {
  constructor(private readonly logger: LoggerService) {}

  intercept(context: ExecutionContext, next: CallHandler): Observable<any> {
    let req: any;
    // @ts-expect-error graphql的情况
    if (context.getType() === 'http' || context.getType() === 'graphql') {
      const args = context.getArgs();
      if (context.getType() === 'http') {
        req = args[0];
      } else {
        req = args[2].req;
      }

      const { body, query, params, raw, method, headers } = req;
      const url = raw.originalUrl || raw.url;

      if (!raw || IGNORE_URL.includes(url)) {
        return next.handle();
      }
      this.logger.dtlog.setCategory('web');
      this.logger.dtlog.set({ url });
      this.logger.log({ method, body, query, params, headers });
      this.logger.dtlog.setCategory('');

      const start = process.hrtime();
      const response = context.switchToHttp().getResponse();
      return next.handle().pipe(
        tap(() => {
          const diff = process.hrtime(start);
          const ms = diff[0] * 1e3 + diff[1] * 1e-6;
          typeof response.headers === 'function' &&
            response.headers({ 'X-Response-Time': ms });
          if (ms > 10000) {
            // 大于10秒打印内存和cpu
            this.logger.info('memoryUsage', process.memoryUsage());
            this.logger.info('cpuUsage', process.cpuUsage());
          }
        }),
      );
    } else {
      return next.handle();
    }
  }
}
