import * as dotenv from 'dotenv';
dotenv.config({
  path: './config/.env',
});

import { NestFactory } from '@nestjs/core';
import { AppModule } from './app.module';
import { FastifyAdapter } from '@nestjs/platform-fastify';
import { ENV_NAME_ENUM } from './common/lib/enum';
import { AllExceptionsFilter } from '@common/filter/all-exceptions.filter';
import { LoggingInterceptor } from '@common/logger/logging.interceptor';

const PORT = process.env.NODE_ENV === 'container' ? 8080 : 4567;

async function bootstrap() {
  const app = await NestFactory.create(
    AppModule,
    new FastifyAdapter({
      bodyLimit: 1048576 * 2, // 最大 Post数据
    }),
  );
  if (process.env.ENV_NAME !== ENV_NAME_ENUM.PORDUCT) {
    app.enableCors();
  }

  app.useGlobalInterceptors(app.get(LoggingInterceptor));
  app.useGlobalFilters(app.get(AllExceptionsFilter));

  await app.listen(PORT, '0.0.0.0');
  console.log(`Application is running on: ${await app.getUrl()}`);
}

bootstrap();
